{
  ZPAQ's Face v1.0 Data Compression

    Copyright (C) 2020 Alexander Staidl
    v1.0: Thanks to the members of the german lazarusforum for support.
          Special thanks to Andy for Mac-Support and Winfried for Bugtracking

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 3 as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, process, Forms, Controls, Graphics, Dialogs, StdCtrls, math, LCLType,
  ExtCtrls, ComCtrls, FileCtrl, EditBtn, ShellCtrls, FileUtil, LazFileUtils, StrUtils
  {$IFDEF UNIX},baseunix{$ENDIF UNIX};                                            //baseunix existiert nur für Unix-Systeme.

type

  { THauptfenster }

  THauptfenster = class(TForm)
    ApplicationProperties1: TApplicationProperties;
    ArchivOeffnenButton: TButton;
    Experteneinstellung: TCheckBox;
    ThreadAnzahl: TComboBox;
    SpeicherBedarf: TComboBox;
    NeuesArchiv: TButton;
    EntpackenButton: TButton;
    Label2: TLabel;
    OrdnerauswahlButton: TButton;
    DateiauswahlButton: TButton;
    Ausgabeaktualisierung: TTimer;
    Label1: TLabel;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    Archivinhalt: TShellListView;
    //Zum Archivinhalt: Dieser ist wie folgt aufgebaut.
    //Archivinhalt.Items.Item[n].Caption enthält: Datei/Ordnernamen + Pfad, Archivinhalt.Items.Item[n].SubItems.Strings[0] enthält Dateigröße, Archivinhalt.Items.Item[n].SubItems.Strings[1] enthält Dateiendung, Archivinhalt.Items.Item[n].SubItems.Strings[2] enthält den Ordner-/Dateinamen OHNE Pfad, Archivinhalt.Items.Item[n].SubItems.Strings[3] enthält 'archviert' falls die Datei/Ordner bereits komprimiert ist.
    SelectDirectoryDialog1: TSelectDirectoryDialog;
    ZPAQ: TProcess;
    StartButton: TButton;
    Kompressionslevel: TComboBox;
    AusgabeMemo: TMemo;
    procedure ApplicationProperties1DropFiles(Sender: TObject;
      const FileNames: array of String);
    procedure ApplicationProperties1EndSession(Sender: TObject);
    procedure ArchivinhaltKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure ArchivOeffnenButtonClick(Sender: TObject);
    procedure AusgabeaktualisierungTimer(Sender: TObject);
    procedure ExperteneinstellungChange(Sender: TObject);
    procedure DateiauswahlButtonClick(Sender: TObject);
    procedure EntpackenButtonClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure KompressionslevelChange(Sender: TObject);
    procedure NeuesArchivClick(Sender: TObject);
    procedure OrdnerauswahlButtonClick(Sender: TObject);
    procedure StartButtonClick(Sender: TObject);
    function DateinameOhnePfad(DateinameMitPfad:string): string;                  //Folgende Funktion streift den Dateipfad ab. Kann auch auf Ordner angewendet werden:
    procedure ThreadAnzahlChange(Sender: TObject);
    procedure ZPAQAusgabeEinlesen(TeilweiseEinlesen:boolean);                     //Liest die Ausgabe von ZPAQ in die AusgabeStrLst ein.
    procedure ArchivinhaltAusAusgabeParsen;                                       //Parst die zpaq-Ausgabe beim Öffnen eines Archivs und speichert de Inhalt in die AusgabeStrLst.
    procedure SpeicherbedarfBerechnen(KLevel, ThreadAnz:integer);                 //KLevel=Kompressionslevel
    procedure LetzteAusgabenzeilenInsMemo(Anzahl: integer);                       //Überträgt die letzten "Anzahl" Zeilen aus der zpaq-Ausgabepipe in das Ausgabememo.
  private

  public

  end;

var
  Hauptfenster: THauptfenster;
  GeoeffnetesArchiv, GeoeffneterOrdner: string;                                   //GeoeffnererOrdner enthält den zuletzt über die Öffnen/hinzufügen-Diagloge genutzten Ordner. GoeffnetesArchiv enthält Pfad + Dateiname eines über den ArchivOeffnenButton geöffneten Archivs.
  AusgabeStrLst: TStringList;                                                     //In die AusgabeStrLst wird die zpaq-Ausgabepipe gespeichert.

implementation

{$R *.lfm}

{ THauptfenster }

procedure THauptfenster.FormCreate(Sender: TObject);
begin
  ZPAQ:=TProcess.Create(nil);                                                     //Erzeugt den Prozess für zpaq.
  //ZPAQ.Executable:= FindDefaultExecutablePath('zpaq');                          //Kann verwendet werden, falls jemand eine im System installierte zpaq-Version verwenden möchte. Ist standardmäßig auskommentiert.
  ZPAQ.Executable:=ProgramDirectory + 'bin' + DirectorySeparator + 'zpaq'         //Nutzt die im Unterordner "bin" dieser Anwendung befindliche Version von ZPAQ. DirectorySeparator='/' (Linux) bzw. '\' (Windows)
                   {$IFDEF Windows} +'64.exe' {$ENDIF Windows}                    //Unter Windows heißt das Programm zpaq64.exe
                   {$IFDEF DARWIN} + 'mac' {$ENDIF};                              //Unter MacOsX heißt das Programm zpaqmac
  {$IFDEF UNIX} fpChmod(ZPAQ.Executable,&777);  {$ENDIF UNIX}                     //Markiert die zpaq-Binärdatei als ausführbar. Nur unter Unix notwendig, da es dieses Dateiattribut unter Windows nicht gibt.
  ZPAQ.Options := [poUsePipes];                                                   //Für das Auslesen der ZPAQ-Ausgabe ist das Setzen dieser Option notwendig.
  GeoeffnetesArchiv:='';                                                          //Zu Beginn ist kein Archiv geöffnet.
  GeoeffneterOrdner:='';                                                          //Zu Beginn ist kein Ordner geöffnet.
  AusgabeStrLst:= TStringList.Create;                                             //Initialisierung der AusgabeStrLst, in welche die zpaq-Outputpipe gespeichert wird.
  AusgabeStrLst.Clear;
  {$IFDEF DARWIN} NeuesArchiv.Enabled:=true; {$ENDIF}                             //Ein Workaround für MacOsx: Nach den Angaben von Andy funktioniert das Komprimieren von Dateien nur, wenn zuvor auf NeuesArchiv geklickt wurde.
end;

procedure THauptfenster.KompressionslevelChange(Sender: TObject);
begin
  If Experteneinstellung.Checked then
    SpeicherbedarfBerechnen(Kompressionslevel.ItemIndex,ThreadAnzahl.ItemIndex);  //Der Speicherbedarf ist abhängig vom Kompressionslevel und wird hier entsprechend angepasst
end;

procedure THauptfenster.NeuesArchivClick(Sender: TObject);
begin
  GeoeffnetesArchiv:='';                                                          //Zurücksetzen des geöffneten Archivs
  Label1.Caption:='Archivinhalt:';                                                //Zusücksetzen der Beschriftung
  Archivinhalt.Clear;                                                             //Leeren des Archivinhalts
  EntpackenButton.Enabled:=false;                                                 //Buttons werden deaktiviert, welche bei einem neuen Archiv keine Bedeutung haben.
  StartButton.Enabled:=false;
  NeuesArchiv.Enabled:=false;
end;

procedure THauptfenster.OrdnerauswahlButtonClick(Sender: TObject);
var NeuerEintrag: TListItem;
    n: Integer;
begin
  If GeoeffneterOrdner<>'' then SelectDirectoryDialog1.InitialDir:=GeoeffneterOrdner;  //Sorgt dafür, dass der Öffnen-Dialog mit dem zuletzt geöffneten Ordner startet.
  If SelectDirectoryDialog1.Execute then for n:=0 to SelectDirectoryDialog1.Files.Count-1 do begin  // SelectDirectoryDialog1.Execute öffnet den Ordner-Auswahldialog und liest die ausgewählten Ordner ein...
    NeuerEintrag:=TListItem.Create(Archivinhalt.Items);                           //Neuer Eintrag in den Archivinhalt (=TShellListView).
    NeuerEintrag.Caption:=SelectDirectoryDialog1.Files.Strings[n];                //Pfad + Ordnername als Caption in die 1. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add('-');                                               //Die Ordnergröße wird nicht angegeben, daher ein '-' in die 2. Spalte des Archivinhalts. ---> ggf. zu späteren Zeitpunkt noch ändern.
    NeuerEintrag.SubItems.Add('Ordner');                                          //Als Typ nun der Eintrag 'Ordner' in die 3. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(DateinameOhnePfad(NeuerEintrag.Caption));           //In die später nicht sichtbare 4. Spalte des Archivinhalts kommt der Ordnername ohne Pfad, damit er später ohne Pfad in das Archiv gespeichert wird.
    NeuerEintrag.SubItems.Add('');                                                //Hier steht '', da der Ordner noch nicht komprimiert im Archiv vorliegt.
    Archivinhalt.Items.AddItem(NeuerEintrag);                                     //Jetzt wird der neue Eintrag dem Archivinhalt hinzugefügt:
    GeoeffneterOrdner:=leftstr(NeuerEintrag.Caption,length(NeuerEintrag.Caption)-length(NeuerEintrag.SubItems.Strings[2]));  //Speichert den Pfad des geöffneten Ordners in die Variable GeoeffneterOrdner
    EntpackenButton.Enabled:=false;                                               //Entpacken macht keinen Sinn, wenn Dateien/Ordner hinzugefügt wurden.
    StartButton.Enabled:=true;                                                    //Mit hinzugefügter Datei/Ordner kann das Archiv "erweitert" werden.
    NeuesArchiv.Enabled:=true;                                                    //Der Nutzer kann mit diesem Button alles wieder zurücksetzen.
  end;
end;

procedure THauptfenster.StartButtonClick(Sender: TObject);
var n:integer;
begin
  If ZPAQ.Running then begin                                                      //D.h.: "Stopp!" wurde gedrückt. ZPAQ wird also nachfolgend angehalten.
    Ausgabeaktualisierung.Enabled:=false;                                         //Timer wird ausgestellt
    ZPAQ.Terminate(0);                                                            //ZPAQ wird angehalten
    EntpackenButton.Enabled:=false;                                               //EndpackenButton wird deaktiviert, da das Archiv vermutlich wegen der Unterbrechung defekt ist.
    ArchivOeffnenButton.Enabled:=true;                                            //Das Öffnen eine Archivs soll wieder möglich sein.
    StartButton.Caption:='Archiv speichern...';                                   //StartButton wird wieder umbenannt.
    Screen.Cursor:=crDefault;                                                     //Mauszeiger wird zurückgesetzt.
  end else if Archivinhalt.Items.Count>0 then begin                               //Falls zpaq nicht schon läuft und das Archiv auch nicht leer ist, beginnt der Speichervorgang.
    //SAVEDIALOG ANSTEUERN:
    If GeoeffnetesArchiv='' then begin                                            //Globale Variable GeoeffnetesArchiv beinhaltet den Pfad und Dateinamen des angezeigten Archivinhalts.
      If GeoeffneterOrdner<>''then SaveDialog1.InitialDir:=GeoeffneterOrdner;     //Legt den Ordner fest, den der SaveDialog inital zeigt.
      If Archivinhalt.Items.Count=1 then SaveDialog1.FileName:=Archivinhalt.Items.Item[0].SubItems.Strings[2] + '.zpaq' //Falls im Archivinhalt nur 1 Datei/Ordner ist, soll im Save-dialog der Name der Datei/Ordner mit .zpaq-Endung bereits vorgegeben werden.
        else SaveDialog1.FileName:='Unbenanntes_Archiv.zpaq';                     //Falls mehrere Dateien/Ordner im Archivinhalt sind, schlägt der SaveDialog diesen Dateinamen vor.
      If SaveDialog1.Execute then GeoeffnetesArchiv:=SaveDialog1.FileName;        //Nach dem Speichern hat das neue Archiv einen Namen.
    end;

    If GeoeffnetesArchiv<>'' then begin                                           //Unter der Voraussetzung, dass das Archiv einen Namen hat, wird nun der Archivierungsvorgang initiiert.
      AusgabeStrLst.Clear;                                                        //Löscht den Inhalt der AusgabeStrLst

      //ZPAQ ANSTEUERUNG:
      With ZPAQ.Parameters do begin                                               //Jetzt folgt die Hinzufügung der Programmparameter:
        Clear;                                                                    //Die Parameter, die ev. noch vorhanden sind, werden geleert:
        add('a');                                                                 //Der Parameter a (=add) sagt ZPAQ, es soll die folgenden Inhalte dem Archiv hinzufügen:
        add(GeoeffnetesArchiv);                                                   //Anschließend wird der Name des Archivs an ZPAQ übergeben:
        for n:=0 to Archivinhalt.Items.Count-1 do                                 //Hier wird ZPAQ die Liste der zu komprimierenden Inhalte übergeben:
          if not (Archivinhalt.Items.Item[n].SubItems.Strings[3]='archiviert') then //Dateien, die bereits komprimiert im Archiv enthalten sind, werden nicht noch einmal hinzugefügt.
            add(Archivinhalt.Items.Item[n].Caption);                              //Übergabe des nicht bereits archivierten Datei/Ordnernamens an zpaq.
        add('-to');                                                               //Nun wird ZPAQ angewiesen, diese Dateien/Ordner OHNE Pfad im internen Verzeichnis abzuspeichern. Dabei befinden sich die Namen ohne Pfad bereits im 3. SubItem des Archivinhalts:
        for n:=0 to Archivinhalt.Items.Count-1 do
          if not (Archivinhalt.Items.Item[n].SubItems.Strings[3]='archiviert') then //Dateien, die bereits komprimiert im Archiv enthalten sind, werden nicht noch einmal hinzugefügt.
            add(Archivinhalt.Items.Item[n].SubItems.Strings[2]);                  //[2], da ab 0 gezählt wird: Hier wird zpaq gesagt, dass es die Datei OHNE absolutem Pfad speichern soll. Die Bezeichnung ist im Strings[2] gespeichert.
        If not Experteneinstellung.Checked then add('-m' + IntToStr(Kompressionslevel.ItemIndex)) else begin //Schließlich wird das Kompressionslevel übermittelt.
          add('-m' + IntToStr(Kompressionslevel.ItemIndex) + IntToStr(SpeicherBedarf.ItemIndex));  //Experteneinstellung: BlockSize wird zusätzlich an ZPAQ übermittelt. -m56 für Methode 5 und BlockSize 6.
          add('-t' + IntToStr(ThreadAnzahl.ItemIndex));                           //Experteneinstellung: Threadanzahl wird zusätzlich übermittelt.
        end; //Experteneinstellung
        {$IFNDEF DARWIN} add('-s1'); {$ENDIF}                                     //Diese Option beeinflusst die Ausgabe von zpaq, dass nur noch der Fortschritt in % und die verbleibende Zeit ausgegeben werden. Funktioniert unter Mac OSX laut Andy nicht so gut, daher das Conditional bzgl darwin.
      end; //with Parameters
      ZPAQ.Execute;                                                               //Ausführen von ZPAQ mit den obigen Parametern:

      AusgabeStrLst.Clear;                                                        //Leeren der AusgabeStrLst, denn diese wird durch die zpaq-Ausgabe jetzt neu gefüllt.
      Label1.Caption:='Archivinhalt: ' + GeoeffnetesArchiv;                       //Dateiname wird im Label oben angezeigt.
      EntpackenButton.Enabled:=false;                                             //Bleibt während des Komprimierens deaktiviert, wird erst nach Abschluss des Komprimiervorgangs wieder aktiviert-
      ArchivOeffnenButton.Enabled:=false;                                         // genauso.
      StartButton.Caption:='STOPP!';                                              //Der Komprimiervorgang soll durch erneuten Klick auf die Schaltfläche angehalten werden können, hier die Umbenennung:
      Screen.Cursor:=crHourGlass;                                                 //Mauszeiger wird in ein Uhrenglas umgewandelt (bei neueren Systemen als Kreis)

      Ausgabeaktualisierung.Enabled:=true;                                        //Der Timer Ausgabeaktualisierung aktualisiert das Fenster für die Konsolenausgabe:
    end else showmessage('Leerer Dateiname - das Archiv konnte nicht gespeichert werden.');
  end;
end; //StartButtonClick

procedure THauptfenster.FormClose(Sender: TObject; var CloseAction: TCloseAction
  );
begin
  If ZPAQ.Running then ZPAQ.Terminate(0);                                         //ZPAQ wird abgewürgt, falls es noch läuft.
  ZPAQ.Free;                                                                      //Freigeben des TProcess.
  AusgabeStrLst.Free;                                                             //Freigabe der Liste, in der die zpaq-Ausgabe gespeichert wird.
end;

procedure THauptfenster.AusgabeaktualisierungTimer(Sender: TObject);
begin
  ZPAQAusgabeEinlesen(true);                                                      //Einlesen der ZPAQ-Ausgabe in die AusgabeStrLst. Der Parameter "true" gibt an, dass ZPAQAusgabeEinlesen nicht wartet, bis zpaq.running=false.
  If not ZPAQ.Running then begin                                                  //Falls ZPAQ fertig komprimiert hat...
    Ausgabeaktualisierung.Enabled:=false;                                         //Timer wird deaktiviert
    StartButton.Enabled:=false;                                                   //Da das Archiv bereits komprimiert wurde, wird der Button erst wieder mit dem Hinzufügen neuer Dateien aktiviert.
    StartButton.Caption:='Archiv speichern...';                                   //StartButton wird zurückgesetzt
    EntpackenButton.Enabled:=true;                                                //Archiv kann also auch direkt wieder entpackt werden...
    EntpackenButton.Caption:='Archiv entpacken...';                               //EntpackenButton wird zurückgesetzt.
    ArchivOeffnenButton.Enabled:=true;                                            //Das Öffnen eines Archivs ist jetzt wieder möglich.
    Screen.Cursor:=crDefault;                                                     //Mauszeiger wird zurückgesetzt.
  end;
  LetzteAusgabenzeilenInsMemo(10);                                                //Nur die letzten 10 Zeilen der zpaq-Ausgabe werden in das Memo geladen, damit durch die Wiedergabe keine unnötigen Systemressourcen gefressen werden.
  application.ProcessMessages;                                                    //Tipp von winni. Ergab bei mir keine Änderung, funktioniert allerdings bei Forennutzer winni unter KDE.
end;

procedure THauptfenster.ExperteneinstellungChange(Sender: TObject);
begin
  If Experteneinstellung.Checked then begin                                       //Wenn der Haken bei Experteneinstellung gesetzt ist, kann man die Threadanzahl und den Speicherbedarf selbst wählen.
    ThreadAnzahl.Enabled := true;                                                 //Je mehr Threads um so schneller...
    SpeicherBedarf.Enabled := true;                                               //... je mehr Speicher man zur Verfügung stellt, um so besser die Kompression.
    SpeicherbedarfBerechnen(Kompressionslevel.ItemIndex,ThreadAnzahl.ItemIndex);  //Die Formel zur Berechnung des Speicherbedarfs habe ich aus den man-Pages von zpaq abgeleitet.
  end else
  begin                                                                           //Falls die Experteneinstellung deaktiviert ist, blassen die ComboBoxen aus.
    ThreadAnzahl.Enabled := false;
    ThreadAnzahl.ItemIndex:=0;                                                    //der Parameter -t0 erzeugt so viele Threads, wie Prozessorkerne existieren.
    SpeicherBedarf.Enabled := false;
    SpeicherBedarf.Clear;
  end;
end;

procedure THauptfenster.ArchivOeffnenButtonClick(Sender: TObject);
var n:integer;
begin
  OpenDialog1.Filter :='Datei(*.zpaq)|*.zpaq';                                    //Dieser Filter stellt sicher, dass im OpenDialog1 nur Dateien mit der Endung *.zpaq angezeigt werden.
  If GeoeffneterOrdner<>'' then OpenDialog1.InitialDir:=GeoeffneterOrdner;        //Sorgt dafür, dass der Öffnen-Dialog mit dem zuletzt geöffneten Ordner startet.
  If not ZPAQ.Running then if OpenDialog1.Execute then                            //Zur Sicherheit: Es kann nur dann ein vorhandenes Archiv geöffnet werden, wenn zpaq nicht am Arbeiten ist.
    if length(OpenDialog1.FileName)>5 then if rightstr(OpenDialog1.FileName,5)='.zpaq' then begin
    AusgabeStrLst.Clear;                                                          //Beim Öffnen spielt die AusgabeStrList eine entscheidende Rolle - hier wird die zpaq-Ausgabe gespeichert und diese enthält alle im Archib enthaltenen Datei- und Ordnernamen.
    Archivinhalt.Clear;                                                           //Archivinhalt wird gelöscht
    With ZPAQ.Parameters do begin
      Clear;                                                                      //Zunächst werden die Parameter bereinigt.
      add('l');                                                                   //Mit der Anweisung 'l' listet ZPAQ den Archivinhalt auf.
      add(OpenDialog1.FileName);                                                  //Übergabe des Dateinamens des zu öffnenden Archivs.
    end;
    ZPAQ.Execute;                                                                 //ZPAQ beginnt mit der Auflistung des Inhalts.

    ZPAQAusgabeEinlesen(false);                                                   //Einlesen der Ausgabe von zpaq. false bedeutet, dass die Ausgabe von ZPAQ vollständig, d.h. bis ZPAQ beendet ist, eingelesen wird.

    ArchivinhaltAusAusgabeParsen;                                                 //Parsen der AusgabeStrLst. Geparster Inhalt wird in dem Archivinhalt gespreichert.

    LetzteAusgabenzeilenInsMemo(100);                                             //Die letzten 100 von zpaq ausgegebenen Zeilen werden (just for fun) ins Memo übertragen.

    GeoeffnetesArchiv:=OpenDialog1.FileName;                                      //Globale Variable beinhaltet Pfad + Namen des geöffneten Archivs.
    GeoeffneterOrdner:=leftstr(GeoeffnetesArchiv,length(GeoeffnetesArchiv)-length(DateinameOhnePfad(GeoeffnetesArchiv)));
    Label1.Caption:='Archivinhalt: ' + GeoeffnetesArchiv;                         //Label1 zeigt den Archivnamen an.
    EntpackenButton.Enabled:=true;                                                //Archiv kann also auch entpackt werden.
    StartButton.Enabled:=false;                                                   //Direkt geöffnetes Archiv muss nicht direkt wieder gespeichert werden.
    NeuesArchiv.Enabled:=true;                                                    //Buttons werden aktiviert/deaktiviert

  end else showmessage('Ausgewählte Datei konnte nicht geöffnet werden. Die Endung muss .zpaq sein.');
  OpenDialog1.Filter :='';                                                        //Zurücksetzen des Dateifilters, da der OpenDialog1 ja auch zum "Datei hinzufügen"-Dialog verwendet wird und hier jede beliebige Datei hinzugefügt werden kann.
end; //ArchivOeffnenButtonClick

procedure THauptfenster.ArchivinhaltKeyDown(Sender: TObject; var Key: Word;       //dient aktuell lediglich dem Löschen markierter Dateien und Ordner, falls DEL oder BACKSPACE gedrückt werden.
  Shift: TShiftState);
var n: integer;
    ArchiviertChecked: boolean;
begin
  n:=0;
  ArchiviertChecked:=false;                                                       //Wird dann true, falls unter den markierten Objekten Dateien/Ordner sind, welche bereits von zpaq archiviert wurden. Diese können nicht nachträglich aus dem Archiv entfernt werden.
  If (ord(Key) in [8,46])and(Archivinhalt.Items.Count>0) then begin  //46=Entf-Taste, 8=Backspace-Taste: Löschen von Objekten aus dem Archivinhalt.
    while n<Archivinhalt.Items.Count do begin                                     //Beginnend bei n=0 wird jeder Eintrag geprüft, ob er.....
      if Archivinhalt.Items.Item[n].Selected then                                 //.....vom Nutzer markiert wurde....
        if not (Archivinhalt.Items.Item[n].SubItems.Strings[3]='archiviert') then begin  //.....bereits archiviert ist.
          Archivinhalt.Items.Item[n].Delete;                                      //Falls der markierte Eintrag noch nicht archiviert wurde, wird er aus dem Archivinhalt gelöscht.
          dec(n);                                                                 //Da es jetzt ein Eintrag weniger ist, muss n nach unten gesetzt werden, damit kein Eintrag übersprungen wird.
        end else ArchiviertChecked:=true;                                         //Falls einer der markierten Einträge bereits archiviert ist, wird unten eine Info dazu gezeigt.
      inc(n)                                                                      //...ein Eintrag weiter...
    end;
  end; // if (ord(Key) in [8,46])
  If ArchiviertChecked=true then showmessage('Unter den markierten Objekten befinden sich bereits archivierte Objekte, die nicht gelöscht werden können.');  //Ausgabe, falls ein bereits archivierter Eintrag gelöscht werden sollte.
end; //ArchivinhaltKeyDown

procedure THauptfenster.ApplicationProperties1DropFiles(Sender: TObject;
  const FileNames: array of String);
var
  i, len:Integer;
  NeuerEintrag: TListItem;
begin
  len:=Length(FileNames);
  for i:=0 to len -1 do begin
    NeuerEintrag:=TListItem.Create(Archivinhalt.Items);                           //Neuer Eintrag in den Archivinhalt (=TShellListView). Frage: Free notwendig oder automatisch mit ShellListView?
    NeuerEintrag.Caption:=FileNames[i];                           //Pfad + Dateiname als Caption in die 1. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(FloatToStr(FileSize(NeuerEintrag.Caption)/1000) + ' kB');  //Dateigröße in kB als SubItem für die 2. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(rightstr(NeuerEintrag.Caption,length(NeuerEintrag.Caption)-length(ExtractFileNameWithoutExt(NeuerEintrag.Caption))));  //Dateiendung als SubItem für die 3. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(DateinameOhnePfad(NeuerEintrag.Caption));           //In die später nicht sichtbare 4. Spalte des Archivinhalts kommt der Dateiname ohne Pfad, damit sie später ohne Pfad in das Archiv gespeichert wird.
    NeuerEintrag.SubItems.Add('');                                                //Hier steht '', da die Datei noch nicht komprimiert im Archiv vorliegt.
    Archivinhalt.Items.AddItem(NeuerEintrag);                                     //Jetzt wird der neue Eintrag dem Archivinhalt hinzugefügt:
    GeoeffneterOrdner:=leftstr(NeuerEintrag.Caption,length(NeuerEintrag.Caption)-length(NeuerEintrag.SubItems.Strings[2]));  //Streift über leftstr den Ordner der neu geöffneten Datei ab.
    EntpackenButton.Enabled:=false;                                               //Entpacken macht keinen Sinn, wenn Dateien/Ordner hinzugefügt wurden.
    StartButton.Enabled:=true;                                                    //Mit hinzugefügter Datei/Ordner kann das Archiv "erweitert" werden.
    NeuesArchiv.Enabled:=true;                                                    //Und natürlich darf der Nutzer auch alles zurücksetzen, wenn wer möchte.
  end;

end;

procedure THauptfenster.ApplicationProperties1EndSession(Sender: TObject);
begin

end;

procedure THauptfenster.DateiauswahlButtonClick(Sender: TObject);
var NeuerEintrag: TListItem;
    n: Integer;
begin
  If GeoeffneterOrdner<>'' then OpenDialog1.InitialDir:=GeoeffneterOrdner;        //Sorgt dafür, dass der Öffnen-Dialog mit dem zuletzt geöffneten Ordner startet.
  If OpenDialog1.Execute then for n:=0 to OpenDialog1.Files.Count-1 do begin      //Öffnet den Dateiauswahldialog. Die For-Schleife lässt n durch alle markierten Dateien wandern.
    NeuerEintrag:=TListItem.Create(Archivinhalt.Items);                           //Neuer Eintrag in den Archivinhalt (=TShellListView). Frage: Free notwendig oder automatisch mit ShellListView?
    NeuerEintrag.Caption:=OpenDialog1.Files.Strings[n];                           //Pfad + Dateiname als Caption in die 1. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(FloatToStr(FileSize(NeuerEintrag.Caption)/1000) + ' kB');  //Dateigröße in kB als SubItem für die 2. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(rightstr(NeuerEintrag.Caption,length(NeuerEintrag.Caption)-length(ExtractFileNameWithoutExt(NeuerEintrag.Caption))));  //Dateiendung als SubItem für die 3. Spalte des Archivinhalts:
    NeuerEintrag.SubItems.Add(DateinameOhnePfad(NeuerEintrag.Caption));           //In die später nicht sichtbare 4. Spalte des Archivinhalts kommt der Dateiname ohne Pfad, damit sie später ohne Pfad in das Archiv gespeichert wird.
    NeuerEintrag.SubItems.Add('');                                                //Hier steht '', da die Datei noch nicht komprimiert im Archiv vorliegt.
    Archivinhalt.Items.AddItem(NeuerEintrag);                                     //Jetzt wird der neue Eintrag dem Archivinhalt hinzugefügt:
    GeoeffneterOrdner:=leftstr(NeuerEintrag.Caption,length(NeuerEintrag.Caption)-length(NeuerEintrag.SubItems.Strings[2]));  //Streift über leftstr den Ordner der neu geöffneten Datei ab.
    EntpackenButton.Enabled:=false;                                               //Entpacken macht keinen Sinn, wenn Dateien/Ordner hinzugefügt wurden.
    StartButton.Enabled:=true;                                                    //Mit hinzugefügter Datei/Ordner kann das Archiv "erweitert" werden.
    NeuesArchiv.Enabled:=true;                                                    //Und natürlich darf der Nutzer auch alles zurücksetzen, wenn wer möchte.
  end;
end;

procedure THauptfenster.EntpackenButtonClick(Sender: TObject);
var n: integer;
    NurMarkierteDateienEntpacken: byte;                                           //Die Variable merkt sich den Nutzerwillen, ab das gesamte Archiv oder nur die darin markierten Dateien/Ordner entpackt werden sollen.
                                                                                  //Es gilt folgende Festlegung: 0=undefiniert, 1=ja, 2=nein
begin                                                                             //0 bedeutet dabei undefiniert: D.h. es ist nicht bekannt, ob nur die markierten Dateien oder das komplette Archiv entpackt werden sollen.
  NurMarkierteDateienEntpacken:=0;                                                //Zunächst ist es nicht bekannt, ob nur markierte Dateien komprimiert werden sollen.
  If GeoeffneterOrdner<>'' then SelectDirectoryDialog1.InitialDir:=GeoeffneterOrdner;   //Sorgt dafür, dass der Öffnen-Dialog mit dem zuletzt geöffneten Ordner startet.
  If ZPAQ.Running then begin                                                      //Falls zpaq gerade am Entpacken ist, kann durch einen erneuten Klick auf den EntpackenButton der Vorgang wieder gestoppt werden.
    Ausgabeaktualisierung.Enabled:=false;                                         //Timer wird ausgestellt
    ZPAQ.Terminate(0);                                                            //ZPAQ wird angehalten
    ArchivOeffnenButton.Enabled:=true;                                            //Das Öffnen eine Archivs soll wieder möglich sein.
    EntpackenButton.Caption:='Archiv entpacken...';                               //EntpackenButton wird wieder umbenannt.
    Screen.Cursor:=crDefault;                                                     //Der Mauszeiger wird wieder Normal anstelle der Sanduhr.
  end else if (GeoeffnetesArchiv<>'')  then                                       //Das Entpacken geht nur, wenn das Archiv geladen ist und zpaq nicht läuft.
    if SelectDirectoryDialog1.Execute then begin                                  //Ordner zum Entpacken wird gewählt
    GeoeffneterOrdner:=SelectDirectoryDialog1.FileName + DirectorySeparator;      //Die Variable für den zuletzt geöffneten Ordner wird auf das Zielverzeichnis gesetzt. DirectoryDeparator hängt noch ein "/" an.

    //ZPAQ Ansteuerung:
    With ZPAQ.Parameters do begin
      clear;                                                                      //ZPAQ-Parameter zurücksetzen
      add('x');                                                                   //Komando x lässt das Archiv extrahieren
      add(GeoeffnetesArchiv);                                                     //Übergabe des Pfads und der Datei des zu entkomprimierenden Archivs
      add('-to');                                                                 //Befehl, wohin das Archiv gespeichert werden soll
      add(SelectDirectoryDialog1.FileName);                                       //Filename enthält den Pfad des Ordners, in den das Archiv entpackt werden soll.
    end;

    //Regelt, ob nur die in der Liste markierten Dateien entpackt werden:
    n:=0;
    while (n<Archivinhalt.Items.Count)and(NurMarkierteDateienEntpacken<>2) do     //Geht die Einträge im Archivinhalt durch. Die Whileschleife wird abgebrochen, falls der Nutzer das gesamte Archiv entpacken will (=2)
    begin
      If Archivinhalt.Items.Item[n].Selected then begin                           //Stellt fest, ob die Datei/der Ordner n vom Nutzer markiert wurde.
        If NurMarkierteDateienEntpacken=0 then if Application.MessageBox('Sollen ausschließlich die markierten Dateien entpackt werden?', 'MessageBoxDemo', MB_YESNO) = IDYES then begin //Fragt nach, ob nur markierte Dateien entpackt werden sollen.
          NurMarkierteDateienEntpacken:=1;                                        //1 bedeutet, es sollen nur markierte Dateien entpackt werden.
          ZPAQ.Parameters.Add('-only');                                           //Der zpaq-Befehl "-only" sagt zpaq, es soll nur die nachfolgenden Dateien entpacken.
        end else NurMarkierteDateienEntpacken:=2;                                 //2 bedeutet, dass das gesamte Archiv entpackt werden soll. In diesem Fall werden keine weiteren Parameter an zpaq übergeben.
        If NurMarkierteDateienEntpacken=1 then                                    //Falls nur markierte Dateien entpackt werden sollen(=1), werden nun die markierten Dateien an zpaq übergeben.
          ZPAQ.Parameters.Add(Archivinhalt.Items.Item[n].Caption);
      end;
      inc(n);                                                                     //...der nächste Eintrag in der Liste....
    end; //while

    ZPAQ.Execute;                                                                 //Entpackvorgang wird gestartet

    EntpackenButton.Caption:='ENTPACKEN STOPPEN!';                                //Damit Nutzer sieht, dass das Programm arbeitet und er die Möglichkeit hat, den Vorgang zu stoppen.
    ArchivOeffnenButton.Enabled:=false;                                           //Während des Entpackens kann kein neues Archiv geöffnet werden.
    Screen.Cursor:=crHourGlass;                                                   //Der Mauszeiger wird zur Sanduhr bis der Entpackvorgang endet.
    Ausgabeaktualisierung.Enabled:=true;                                          //Der Output von ZPAQ wird in die Konsole gespeichert. Das geschieht über den Timer Ausgabeaktualisierung im Sekundentakt.
  end;
end; //EntpackenButtonClick

procedure THauptfenster.ThreadAnzahlChange(Sender: TObject);
begin
  SpeicherbedarfBerechnen(Kompressionslevel.ItemIndex,ThreadAnzahl.ItemIndex);    //Die Speicherbedarfoptionen werden angepasst: Da Threads parallel arbeiten ist die Anzahl der Threads proportional zum benötigten Speicherbedarf.
end;


//EIGENE FUNKTIONEN UND PROZEDUREN:

//Folgende Funktion streift den Pfad von Dateien UND Ordnern ab.
function THauptfenster.DateinameOhnePfad(DateinameMitPfad:string): string;        //Diese Prozedur streift den Pfad ab und übrig bleibt der reine Datei- oder Ordnername mit möglicher Endung.
var n: integer;
begin
  RESULT:=DateinameMitPfad;                                                       //Falls die Datei bereits ohne Pfad vorliegt, wird der Rückgabewert im Vorfeld initialisiert als:
  n:=length(DateinameMitPfad)-1;                                                  //Der DateinameMitPfad wird von hinten nach vorne untersucht. Das hinterste Zeichen wird nicht berücksichtigt, damit die Kompatibilität zu Ordnern erhalten bleibt.
  while (n>0) and (DateiNameMitPfad[n]<>DirectorySeparator) do dec(n);            //Das Ende des vorangestellten Pfades wird durch das Symbol '/' erkannt (bzw. unter Windows mit '\'=DirectorySeparator)
  If DateinameMitPfad[n]=DirectorySeparator then                                  //Hier wird nun der Pfad abgestreift, falls er vorhanden ist:
    DateinameOhnePfad:=rightstr(DateinameMitPfad, length(DateinameMitPfad)-n);
end; //DateinameOhnePfad

procedure THauptfenster.ZPAQAusgabeEinlesen(TeilweiseEinlesen:boolean);           //Diese Prozedur ließt den Outputstream von zpaq in die AusgabeStrLst ein.
var
  OutputStream : TStream;                                                         //In diesen Stream wird die Ausgabe von ZPAQ gespeichert.
  NeueZeilen: TStringList;                                                        //Das sind die Zeilen, die nach jedem Auslesen zur AusgabeStrLst hinzugefügt werden.
  BytesRead, VerfuegbareBytes: longint;                                           //Variablen, die speichern, wie groß der Output ist.
  Buffer       : array of byte;                                                   //Hier wird die Ausgabe von ZPAQ zwischengespeichert.

begin
  OutputStream := TMemoryStream.Create;                                           //Outputstream wird als MemoryStream erzeugt.
  NeueZeilen := TStringList.Create;                                               //NeueZeilen ist die Liste mit den neuen Zeilen der ZPAQ-Ausgabe
  repeat
    VerfuegbareBytes:=ZPAQ.Output.NumBytesAvailable;                              //Speichert, wie viele verfügbare Bytes ZPAQ ausgibt.
    setlength(Buffer, VerfuegbareBytes);                                          //Die Puffergröße wird entsprechend gesetzt
    BytesRead := ZPAQ.Output.Read(Buffer[0], VerfuegbareBytes);                   //Die ZPAQ-Ausgabe wird in den Buffer übertragen...
    OutputStream.Write(Buffer[0], BytesRead);                                     //...und vom Buffer in den OutPutStream übertragen.
  until (BytesRead = 0)and((ZPAQ.Running=false) or (TeilweiseEinlesen=true));     //Falls TeilweiseEinlesen=false, dann läuft die Schleife so lange, bis ZPAQ beendet ist.
  OutputStream.Position:=0;                                                       //Der OutputStream soll anschließend beginnend von Position 0 übergeben werden.
  NeueZeilen.LoadFromStream(OutputStream);                                        //Der OutputStream wird in NeueZeilen gespeichert
  AusgabeStrLst.AddStrings(NeueZeilen);                                           //Die neuen Zeilen werden der AusgabeStrLst hinzugefügt

  OutputStream.Free;                                                              //Freigabe des OutputStream.
  NeueZeilen.Free;                                                                //Freigabe von NeueZeilen
end;

procedure THauptfenster.ArchivinhaltAusAusgabeParsen;                             //Diese Prozedur parst die in der AusgabeStrLst gespeicherte Ausgabe von zpaq beim Auflisten des Inhalts des zu öffnenden Archivs.
var Zeile, n: integer;
    ZwSpStr: string;
    NeuerEintrag: TListItem;                                                      //NeuerEintrag ist ein neuer Eintrag in der angezeigten Dateiliste des Archivinhalts.
begin
  Zeile:=0;                                                                       //Die Ausgabe von zpaq befindet sich in der AusgabeStrLst. Dieser wird beginnend mit der Zeile Null ausgelesen.
  ZwSpStr:='';
  while Zeile<AusgabeStrLst.Count do begin                                        //Zeile für Zeile wird die AusgabeStrLst geparst.
    If length(AusgabeStrLst.Strings[Zeile])>0 then                                //Manche Zeilen sind Leerzeilen, die nicht geparst werden sollen.
      if (AusgabeStrLst.Strings[Zeile][1]='-')and(length(AusgabeStrLst.Strings[Zeile])>42) then begin   //ZPAQ-Ausgabezeilen mit Dateiauflistung beginnen mit einem "-". Um Einlesefehler abzufangen wird noch geprüft, ob mehr als 42 Zeichen in der Zeile sind - ansonsten gibt es unten Zugriffsfehler.
        NeuerEintrag:=TListItem.Create(Archivinhalt.Items);                       //Für einen neuen Eintrag im Archivinhalt
        NeuerEintrag.SubItems.Add('');                                            //Für Dateigröße
        NeuerEintrag.SubItems.Add('');                                            //Für Dateityp
        NeuerEintrag.SubItems.Add('');                                            //Für Datei ohne Pfad
        NeuerEIntrag.SubItems.Add('archiviert');                                  //Der Eintrag 'archviert' gibt an, dass die Datei bereits im komprimierten Zustand vorliegt.
                                                                                  //Aufbau AusgabeStrLst:Datum Uhrzeit Dateigröße d=Ordner Pfad/DateiOderOrdner.MöglicheEndung  Beispiele:
                                                                                  //Aufbau: "- 2020-07-16 16:27:46       747977 d0775 lib/x86_64-linux/" im Fall eines Ordners: d=Directory
                                                                                  //Aufbau: "- 2020-07-15 10:52:44         2048  0664 /home/alexander/Schreibtisch/ZPAQ-FACE/unit1.lfm" im Fall einer Datei
        ZwSpStr:=midstr(AusgabeStrLst.Strings[Zeile],42, length(AusgabeStrLst.Strings[Zeile]));  //Ab ab dem 42. Zeichen beginnt in der Ausgabe von zpaq immer der Dateiname+Pfad
        NeuerEintrag.Caption:=ZwSpStr;                                            //Pfad + Dateiname werden an den neuen Eintrag übergeben.
        NeuerEintrag.SubItems.Strings[2]:=DateinameOhnePfad(ZwSpStr);             //Nur der Vollständigkeit halber: In nicht sichtbarer 4. Spalte der Tabelle wird die Datei/der Ordner OHNE Pfad gespeichert.
        ZwSpStr:='';

        If AusgabeStrLst.Strings[Zeile][36]='d' then NeuerEintrag.SubItems.Strings[1]:='Ordner'    //Im Fall eines Ordners steht an 36. Stelle ein d. Bei Dateityp wird "Ordner" angezeigt, falls es sich um einen Ordner handelt, ansonsten die Dateiendung...
          else NeuerEintrag.SubItems.Strings[1]:=rightstr(NeuerEintrag.Caption,length(NeuerEintrag.Caption)-length(ExtractFileNameWithoutExt(NeuerEintrag.Caption)));  //Dateiendung falls vorhanden.

        n:=34;                                                                    //Nachfolgend wird von der 34. Stelle nach hinten bis zum ersten Leerzeichen geparst. Hier steht die Dateigröße.
        ZwSpStr:='';                                                              //Es muss von hinten nach vorne geparst werden, da die Dateigröße in der Ausgabe von zpaq rechtsbündig enthalten ist.
        repeat
          ZwSpStr:=AusgabeStrLst.Strings[Zeile][n] + ZwSpStr;                     //Da von hinten nach vorne geparst wird, muss das jeweils neue Zeichen von links aufaddiert werden.
          dec(n)                                                                  //Das nächste Zeichen...
        until (AusgabeStrLst.Strings[Zeile][n]=' ') or (n=1);                     //Es wird bis zum Leerzeichen geparst. D.h. ZwSpStr enthält jetzt die Dateigröße in Bytes.

        NeuerEintrag.SubItems.Strings[0]:=FloatToStr(StrToInt(ZwSpStr)/1000) + ' kB';  //Dateigröße in kB an den neuen Eintrag übergeben.
        Archivinhalt.Items.AddItem(NeuerEintrag);                                 //Neuer Eintrag wird an den Archivinhalt übergeben.

      end; //if AusgabeStrLst.Strings[Zeile][1]='-'
    inc(Zeile);                                                                   //Auf geht's zur nächsten Zeile :)
  end; //while Zeile<AusgabeStrLst.Count
end; //procedure ArchivinhaltAusAusgabeParsen

procedure THauptfenster.SpeicherbedarfBerechnen(KLevel, ThreadAnz:integer);       //Die Prozedur berechnet den erwarteten Speicherbedarf von zpaq und gibt dem Nutzer in der zugehörigen Combobox Auswahlmöglichkeiten.
var BlockSize:integer;                                                            //Je nach Auswahl verwendet zpaq eine andere Blocksize: Je größer, um so besser die Kompressionsdichte.
    Speicher: real;
begin
  SpeicherBedarf.Items.Clear;                                                     //Die ComboBox wird zunächst geleert.
  If ThreadAnz = 0 then begin                                                     //ThreadAnz=0 bedeutet, dass zpaq so viele Threads erzeugt, wie es Prozessorkerne gibt. In diesem Fall kann der...
    for BlockSize:=0 to 11 do                                                     //...zpaq-Blocksize-Parameter selbst gewählt werden.
      SpeicherBedarf.Items.Add('BlockSize: ' + IntToStr(BlockSize));
  end else begin
    for BlockSize:=0 to 11 do begin
      If KLevel = 5 then Speicher := 16 * ThreadAnz * power(2,BlockSize) else     //KLevel ist das vom Nutzer gewählte Kompressionslevel (-m0 bis -m5). Nur bei -m5 ist der benötigte Arbeitsspeicher deutlich größer, was im Faktor 16 ausgedrückt wird.
      Speicher := 8 * ThreadAnz * power(2,BlockSize);                             //Speicherbedarfsberechnung für -m0 bis -m4. Entsprechend den Angaben im Manual von zpaq.
      SpeicherBedarf.Items.Add(FloatToStr(Speicher) + ' MiB');                    //Hinzufügen des prognostizierten Speicherbedarfs in die ComboBox.
    end;
  end;
  SpeicherBedarf.ItemIndex:=0;                                                    //Standardmäßig wird der geringsmögliche Speicherbedarf voreingestellt.
end; //SpeicherbedarfBerechnen

procedure THauptfenster.LetzteAusgabenzeilenInsMemo(Anzahl: integer);             //Die Ausgabe von zpaq wird in AusgabeStrLst gespeichert. Diese Prozedur übergibt die letzten "Anzahl" Zeilen an das Memo zur Information an den Nutzer.
var Anfang, n: integer;                                                           //Anfang enthält die erste Zeile, die an das Memo übergeben wird.
begin
  AusgabeMemo.Clear;                                                              //Das Memo wird geleert.
  If AusgabeStrLst.Count<Anzahl then Anfang:=0 else Anfang := AusgabeStrLst.Count-Anzahl; //Hier wird berechnet, ab welcher Zeile die Übergabe an das Memo erfolgt.
  for n := Anfang to AusgabeStrLst.Count-1 do AusgabeMemo.Lines.Add(AusgabeStrLst.Strings[n]);  //Jetzt werden die letzten "Anzahl" Zeilen der AusgabeStrLst an das AusgabeMemo übergeben.
  AusgabeMemo.SelStart := Length(AusgabeMemo.Text);                               //Das Ausgabememo wird so eingestellt, dass die letzte Zeile gesehen werden kann.
  AusgabeMemo.Refresh;                                                            //Neuen Inhalt anzeigen.
end;  //LetzteAusgabenzeilenInsMemo

end.

